import warnings

warnings.warn("pymatgen.matproj.rest has been moved to pymatgen.ext.matproj."
              "This stub will be removed in pmg 2018.01.01.",
              DeprecationWarning)

from pymatgen.ext.matproj import *

